//
// Bi-Di(RailCom) Sender Rev.exp.01
// Copyright(C)'2020- Nagoden / Desktop Station
//
#include <Arduino.h>
#include <string.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <SoftwareSerial.h>
#include "NmraDcc_x.h"

/*************************************************

   Declaration of Symbols

 *************************************************/

//#define DEBUG			//リリースのときはコメントアウトすること
#define MAN_VER_NUMBER 01   /* Release Ver CV07 */
#define MAN_ID_NUMBER 108  /* Manufacture ID */

#define DECODER_ADDRESS 3
#define CV_28_CONFIG 28
//Pin Func  Port
#define CV_ACK_A 5    //CV ACK Port
#define CV_ACK_B 6    //CV ACK Port



//Codings in the 4 out of 8 code NMRA RCN-217 Table2(0x00 to 0x3F)
//変換表からの逆変換テーブル

const int convtable[64] = {
  172, 170, 169, 165, 163, 166, 156, 154, 153, 149, 147, 150, 142, 141, 139, 177,
  178, 180, 184, 116, 114, 108, 106, 105, 101, 99, 102, 92, 90, 89, 85, 83,
  86, 78, 77, 75, 71, 113, 232, 228, 226, 209, 201, 197, 216, 212, 210, 202,
  198, 204, 120, 23, 27, 29, 30, 46, 54, 58, 39, 43, 45, 53, 57, 51
};

/*************************************************

   Declaration of Classes

 *************************************************/

NmraDcc	 Dcc;
DCC_MSG	 Packet;

/*************************************************

   Declaration of Variables

 *************************************************/


//Task Schedule
//unsigned long gPreviousL1 = 0;

//Definition of internal variables
uint8_t gCutout_on = 0 ;  //カットアウトONフラグ
uint8_t gFour_digit = 0 ; //4桁アドレスフラグ
uint8_t gSend_count = 0 ; //コマンド送信カウント
byte sID1 = B00000000 ;
byte sID2 = B00000000 ;
byte bitwork1 = B00000000;
byte bitwork2 = B00000000;
byte bitwork3 = B00000000;
byte bitwork4 = B00000000;
byte bitworkA = B00000000;
byte bitworkB = B00000000;
byte bitworkC = B00000000;
byte bitworkD = B00000000;

// Variables of CV to use
uint8_t gCV1_SAddr = 3;
uint8_t gCVx_LAddr = 3;
uint8_t gCV17_Addr = 0;
uint8_t gCV18_Addr = 0;
uint8_t gCV28_Vstart = 1;
uint8_t gCV29_Vstart = 10;

struct CVPair {
  uint16_t	CV;
  uint8_t	Value;
};

CVPair FactoryDefaultCVs [] = {
  {CV_MULTIFUNCTION_PRIMARY_ADDRESS, DECODER_ADDRESS},
  {CV_ACCESSORY_DECODER_ADDRESS_MSB, 0},		//The LSB is set CV 1 in the libraries .h file, which is the regular address location, so by setting the MSB to 0 we tell the library to use the same address as the primary address. 0 DECODER_ADDRESS
  {CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB, 0},	 //XX in the XXYY address
  {CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB, 0},	 //YY in the XXYY address
  {CV_28_CONFIG, 1},
  {CV_29_CONFIG, 10}

};

/*************************************************

   Declaration of Functions

 *************************************************/

void(* resetFunc) (void) = 0;  //declare reset function at address 0
uint8_t FactoryDefaultCVIndex = sizeof(FactoryDefaultCVs) / sizeof(CVPair);
void createAdressData(void);
/*************************************************

   Functions

 *************************************************/


/*************************************************
   CV Reset FactoryDefault
 *************************************************/
void notifyCVResetFactoryDefault()
{
  //When anything is writen to CV8 reset to defaults.
  resetCVToDefault();
#ifdef DEBUG
  Serial.println("Resetting...");
#endif
  delay(1000);  //typical CV programming sends the same command multiple times - specially since we dont ACK. so ignore them by delaying
  resetFunc();
}

/*************************************************
   CV Reset
 *************************************************/
void resetCVToDefault()
{
  //CVをデフォルトにリセット
#ifdef DEBUG
  Serial.println("CVs being reset to factory defaults");
#endif
  for (int j = 0; j < FactoryDefaultCVIndex; j++ ) {
    Dcc.setCV( FactoryDefaultCVs[j].CV, FactoryDefaultCVs[j].Value);
  }
}

/*************************************************
   CV Change Message
 *************************************************/
extern void notifyCVChange( uint16_t CV, uint8_t Value)
{
  //CVが変更されたときのメッセージ
#ifdef DEBUG
  Serial.print("CV ");
  Serial.print(CV);
  Serial.print(" Changed to ");
  Serial.println(Value, DEC);
#endif
}

/*************************************************
   CV Ack
 *************************************************/
void notifyCVAck(void)
{
#ifdef DEBUG
  Serial.println("notifyCVAck");
#endif

  //CV_Ack();
  digitalWrite(CV_ACK_B, HIGH);
  digitalWrite(CV_ACK_A, LOW);
  delay( 6 );
  digitalWrite(CV_ACK_B, LOW);
  digitalWrite(CV_ACK_A, LOW);
}

/*************************************************
   setup
 *************************************************/
void setup()
{

  //シリアル通信開始
  Serial.begin (250000);
  // ソフトウエアシリアル通信レートセット:

  //割り当てピン初期化
  //ACK出力ピン D5,D6を出力にセット
  pinMode(CV_ACK_B, OUTPUT);
  pinMode(CV_ACK_A, OUTPUT);

  //デコーダがCV初期値が無い時に初期値セット
  if ( Dcc.getCV(CV_MULTIFUNCTION_PRIMARY_ADDRESS) == 0xFF )
  {
    //if eeprom has 0xFF then assume it needs to be programmed
#ifdef DEBUG
    Serial.println("CV Defaulting due to blank eeprom");
#endif
    notifyCVResetFactoryDefault();
  }
  else {
#ifdef DEBUG
    Serial.println("CV Not Defaulting");
#endif
  }

  // Setup which External Interrupt, the Pin it's associated with that we're using, disable pullup.
  Dcc.pin(0, 2, 0);

  // Call the main DCC Init function to enable the DCC Receiver
  Dcc.init( MAN_ID_NUMBER, MAN_VER_NUMBER,   FLAGS_MY_ADDRESS_ONLY , 0 );

  //Reset task
  //  	    gPreviousL1 = millis();

  //Init CVs
  gCV1_SAddr = Dcc.getCV( CV_MULTIFUNCTION_PRIMARY_ADDRESS ) ;
  gCVx_LAddr = (Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB ) << 8) + Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB );
  gCV17_Addr =  Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB ) ;
  gCV18_Addr =  Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB ) ;
  gCV28_Vstart = Dcc.getCV( CV_28_CONFIG ) ;
  gCV29_Vstart = Dcc.getCV( CV_29_CONFIG ) ;

#ifdef DEBUG
  Serial.print("CV1(ShortAddr): ");
  Serial.println(gCV1_SAddr);
  Serial.print("CV17/18(LongAddr): ");
  Serial.println(gCVx_LAddr);
  Serial.print("CV17(LSBAddr): ");
  Serial.println(gCV17_Addr);
  Serial.print("CV18(MSBAddr): ");
  Serial.println(gCV18_Addr);
  Serial.println("Ready");
#endif
  // Get Send address data
  createAdressData();

  // The D2 pin interrupts the start of Cutout
  attachInterrupt( 1 , cutout, FALLING);
}

/*************************************************
   loop
 *************************************************/
void loop()
{

  if ( (gCV29_Vstart  &  B00001000) == 8 ){  //CV29のビット３が１の場合のみ動作

    if (gCutout_on == 1) {

#ifdef DEBUG
      Serial.println("Recive CUTOUT ");
#endif
      if (gSend_count == 0) {
        Serial.write (bitworkA);
        Serial.write (bitworkB);
        gSend_count = 1 ;
#ifdef DEBUG
        Serial.println("1st send command ");
#endif
      }
      else if  (gSend_count == 1) {
        Serial.write (bitworkC);
        Serial.write (bitworkD);
        gSend_count = 0 ;
#ifdef DEBUG
        Serial.println("2nd send command ");
#endif
      }
      gCutout_on = 0 ;
    }
  }  
      // You MUST call the NmraDcc.process() method frequently from the Arduino loop() function for correct library operation
     Dcc.process();
}


/*************************************************
   Cutout signal reception event
 *************************************************/
//カットアウト開始フラグの処理
void cutout() 
{
  gCutout_on = 1 ;
}

/*************************************************
   Event due to reception of DCC velocity signal
 *************************************************/
//DCC速度信号の受信によるイベント
extern void notifyDccSpeed( uint16_t Addr, DCC_ADDR_TYPE AddrType, uint8_t Speed, DCC_DIRECTION Dir, DCC_SPEED_STEPS SpeedSteps )
{

}

/*************************************************
   Function signal reception event
 *************************************************/
//ファンクション信号受信のイベント
extern void notifyDccFunc( uint16_t Addr, DCC_ADDR_TYPE AddrType, FN_GROUP FuncGrp, uint8_t FuncState)
{

}

/*************************************************
   Send address data creation
 *************************************************/
void createAdressData(void)
{
  //Valid only when 1 is set in Bit3 of CV29
  //CV29のBit3に1がセットされている場合のみ有効
  if ((bitRead(gCV29_Vstart, 3)) == 1) {
    //ID1 ID2 のセット
    sID1 = B00000001;
    sID2 = B00000010;
    //4桁アドレスかCV29のビット5で判定
    if ( (bitRead(gCV29_Vstart, 5)) == 1) {
      gFour_digit = 1 ; //4桁アドレス
      // ID1 + gCV17_Addr
      bitwork1 = sID1 << 2 ;
      bitwork2 =  gCV17_Addr & B10111111;
      bitWrite(bitwork1, 0, bitRead( bitwork2, 6));
      bitWrite(bitwork1, 1, bitRead( bitwork2, 7));
      bitwork2 = gCV17_Addr & B00111111;
      // ID2 + gCV18_Addr
      bitwork3 = sID2 << 2 ;
      bitWrite(bitwork3, 0, bitRead( gCV18_Addr, 6));
      bitWrite(bitwork3, 1, bitRead( gCV18_Addr, 7));
      bitwork4 = gCV18_Addr & B00111111;
    }
    else {
      gFour_digit = 0 ; //2桁アドレス
      // ID1 + 0  0001.00000000
      bitwork1 = sID1 << 2 ;
      bitWrite(bitwork1, 0, 0);
      bitWrite(bitwork1, 1, 0);
      bitwork2 =  B00000000 ;
      // ID2 + gCV1_SAddr　0010.gCV1_SAddr　
      bitwork3 = sID2 << 2 ;
      bitWrite(bitwork3, 0, bitRead( gCV1_SAddr, 6));
      bitWrite(bitwork3, 1, bitRead( gCV1_SAddr, 7));
      bitwork4 = gCV1_SAddr & B00111111;

    }
    //変換テーブルで送信データ作成
    bitworkA = convtable[bitwork1] ;
    bitworkB = convtable[bitwork2] ;
    bitworkC = convtable[bitwork3] ;
    bitworkD = convtable[bitwork4] ;
#ifdef DEBUG
    Serial.print("COMMAND1): ");
    Serial.print(bitworkA, HEX);
    Serial.print("    COMMAND2): ");
    Serial.println(bitworkB, HEX);
    Serial.print("COMMAND3): ");
    Serial.print(bitworkC, HEX);
    Serial.print("    COMMAND4): ");
    Serial.println(bitworkD, HEX);
#endif
  }
}
